import tkinter as Tk
from subprocess import run

TERMINAL_BACKGROUND = '#000000'

class ActiveWindow:
        def __init__(self):
                # use xdotool to query basic info about the terminal's geometry
                out = run(['xdotool','getactivewindow','getwindowgeometry'], capture_output=True)
                # TODO warn user if they do not have xdotool installed
                # parse output and calculate references
                out = str(out.stdout, encoding='utf-8')
                window_id, abs_pos, geometry = out.splitlines()
                self.left, self.top = abs_pos.split()[1].split(',')
                self.left, self.top = int(self.left), int(self.top)
                self.width, self.height = geometry.split()[1].split('x')
                self.width, self.height = int(self.width), int(self.height)
                self.right = self.left + self.width
                self.bottom = self.top + self.height
                self.center_x = self.left + self.width // 2
                self.center_y = self.top + self.height // 2

class DisplayImage:
        """A simple class to draw an image to the screen.
        MUST BE PNG
        There are two methods: show and close.
        It can also be invoked as a context manager."""

        def __init__(self, *, image_path=None, data_bytes=None):
                self.image_path = image_path
                self.data_bytes = data_bytes

        def show(self):
        # get geometry information about the active terminal
                aw = ActiveWindow()

                self.root = Tk.Tk()

                if self.image_path:
                        self.image = Tk.PhotoImage(file=self.image_path)
                elif self.data_bytes:
                        self.image = Tk.PhotoImage(data=self.data_bytes)
                # warning: images cannot be stored as local variables
                # of a function or else there is risk of them being
                # destroyed by premature garbage collection.

                # we want to center the image at the bottom of the terminal
                # with a 40 pixel margin on all sides
                # if the image is too big we will only see part of it
                # but the window will fit inside the active terminal
                img_height = self.image.height()
                img_width = self.image.width()
                margin = 40
                left = max(aw.left + margin, aw.center_x - img_width // 2)
                right = min(aw.right - margin, aw.center_x + img_width // 2)
                width = right - left
                bottom = aw.bottom - margin
                top = max(aw.top + margin, bottom - img_height)
                height = bottom - top
                self.root.geometry(f'{width}x{height}+{left}+{top}')

                # we do not want to let the window manager make decisions
                self.root.overrideredirect(True)

                # draw the image on a canvas
                # place our canvas to occupy the whole window
                self.canvas = Tk.Canvas(self.root, width=width, height=height)
                self.canvas.config(bg=TERMINAL_BACKGROUND)
                self.canvas.create_image(0,0, anchor=Tk.NW, image=self.image)
                self.canvas.place(x=-1,y=-1, height=height+2, width=width+2)

                # manually draw our window to the screen
                # it is common to see root.mainloop(), but
                # that is unneeded as there is not UX here.
                self.root.update()

        def close(self):
                assert self.root
                self.root.destroy()

        def __enter__(self):
                if self.data_bytes or (self.image_path and self.image_path.exists()):
                        self.show()

        def __exit__(self, *exception_args):
                if self.data_bytes or (self.image_path and self.image_path.exists()):
                        self.close()



""" vinca statistics module """

from vinca._lib.julianday import today
from vinca._lib.terminal import COLUMNS
from vinca._lib import unicode_bitmaps
from vinca._lib import ansi

from rich import console
from rich import align
print = console.Console().print

class Statistics:

    def __init__(self, cursor, interval=7):
            self.cursor = cursor
            self.interval = interval
            self.bincount = 100
            self.height = 6

    @property
    def current_week(self):
            return today() // self.interval

    def review_counts(self):
            min_week = self.current_week - self.bincount + 1
            self.cursor.execute('SELECT round(date / ?) as week, count(*) as count FROM reviews'
             ' GROUP BY week HAVING week >= ?', (self.interval, min_week))
            rows = self.cursor.fetchall()
            d = {week: 0 for week in range(min_week, self.current_week)}
            for row in rows:
                week, count = row
                d[week] = count
            return d.values()

    def create_counts(self):
            min_week = self.current_week - self.bincount + 1
            self.cursor.execute('SELECT round(create_date / ?) as week, count(*) as count FROM cards'
             ' GROUP BY week HAVING week >= ?', (self.interval, min_week))
            rows = self.cursor.fetchall()
            d = {week: 0 for week in range(min_week, self.current_week)}
            for row in rows:
                week, count = row
                d[week] = count
            return d.values()

    def counts_to_scores(self, counts):
            score_unit = max(counts) // self.height
            score_unit = max(score_unit, 1)
            return [1 if 0 < count < score_unit else count // score_unit for count in counts]

    def scores_to_bitmap(self, scores):
            l = scores                                          # 4
            l = ['1' * score for score in l]                    # 1111
            l = [f'{score:0>{self.height}s}' for score in l]    # 001111
            l = [[int(d) for d in list(score)] for score in l]  # [0,0,1,1,1,1]
            l = [list(col) for col in zip(*l)]                  # transpose matrix to be horizontal
            return unicode_bitmaps.Bitmap(l)

    def counts_to_unicode(self, counts):
            return self.scores_to_bitmap(self.counts_to_scores(counts)).to_unicode()

    def review_stats(self):
            self.cursor.execute('SELECT count(*) FROM reviews')
            total_reviews = self.cursor.fetchone()[0]
            if total_reviews is None: total_reviews = 0
            self.cursor.execute('SELECT sum(seconds) FROM reviews')
            total_time = self.cursor.fetchone()[0]
            if total_time is None: total_time = 0
            self.cursor.execute('SELECT min(date) FROM reviews')
            first_date = self.cursor.fetchone()[0]
            if first_date is None: first_date = today() - 1
            total_days = today() - first_date
            reviews_per_day = total_reviews / total_days
            time_per_review = total_time / total_reviews if total_reviews else 0
            time_per_day = total_time / total_days
            self.cursor.execute('SELECT count(*) FROM reviews WHERE date > ?',(today() - self.interval,))
            recent_reviews = self.cursor.fetchone()[0]
            self.cursor.execute('SELECT sum(seconds) FROM reviews WHERE date > ?',(today() - self.interval,))
            recent_time = self.cursor.fetchone()[0]
            if recent_time is None: recent_time = 0
            return (f'{total_reviews} reviews '
                    f'{reviews_per_day:.1f} per day '
                    f'{recent_reviews} in the past {self.interval} days\n'
                    f'{time_per_review:.1f} seconds per review {time_per_day // 60:.1f} minutes per day\n'
                    f'{total_time // 3600} hours '
                    f'{recent_time // 60} minutes in the last {self.interval} days')

    def create_stats(self):
            self.cursor.execute('SELECT count(*) FROM cards')
            total_cards = self.cursor.fetchone()[0]
            if total_cards is None: total_cards = 0
            self.cursor.execute('SELECT min(create_date) FROM cards')
            first_date = self.cursor.fetchone()[0]
            if first_date is None: first_date = today() - 1
            total_days = today() - first_date
            cards_per_day = total_cards / total_days
            self.cursor.execute('SELECT count(*) FROM cards WHERE create_date > ?',(today() - self.interval,))
            created_recent = self.cursor.fetchone()[0]
            if created_recent is None: created_recent = 0
            return (f'{total_cards} cards created '
                    f'{cards_per_day:.1f} per day '
                    f'{created_recent} in the past {self.interval} days')

    def print(self):
            review_map = self.counts_to_unicode(self.review_counts())
            create_map = self.counts_to_unicode(self.create_counts())
            print(justify='center')
            print('[underline]STATISTICS',style='bold',end='',justify='center')
            print(f'graphs show {self.bincount} intervals of {self.interval} days',justify='center')
            print(justify='center')
            print(align.Align.center(review_map), style='green')
            print('▔'*(self.bincount//2), style='red',justify='center')
            print(self.review_stats(), style='green',justify='center')
            print(justify='center')
            print(justify='center')
            print(align.Align.center(create_map), style='blue')
            print('▔'*(self.bincount//2), style='red',justify='center')
            print(self.create_stats(), style='blue',justify='center')
            print(justify='center')




from vinca._lib import ansi
import shutil

COLUMNS, LINES = shutil.get_terminal_size()

# A Context Manager for the terminal's alternate screen
class AlternateScreen:
        def __enter__(self):
                ansi.save_cursor()
                ansi.hide_cursor()
                ansi.save_screen()
                ansi.clear()
                ansi.move_to_top()

        def __exit__(self, *exception_args):
                ansi.restore_screen()
                ansi.restore_cursor()


class LineWrapOff:
        def __enter__(self):
                ansi.line_wrap_off()

        def __exit__(self, *exception_args):
                ansi.line_wrap_on()



from pathlib import Path

quads_to_unicode = {((0,0),
 (0,0)):' ',
((0,0),
 (0,1)):'▗',
((0,0),
 (1,0)):'▖',
((0,0),
 (1,1)):'▄',
((0,1),
 (0,0)):'▝',
((0,1),
 (0,1)):'▐',
((0,1),
 (1,0)):'▞',
((0,1),
 (1,1)):'▟',
((1,0),
 (0,0)):'▘',
((1,0),
 (0,1)):'▚',
((1,0),
 (1,0)):'▌',
((1,0),
 (1,1)):'▙',
((1,1),
 (0,0)):'▀',
((1,1),
 (0,1)):'▜',
((1,1),
 (1,0)):'▛',
((1,1),
 (1,1)):'█'}


class Bitmap(list):
        # def __init__(self, l):
        #       list.__init__(self, l)
        #       self.rows = len(self)
        #       self.cols = len(self[0])
        #       assert all((len(row) == self.cols for row in self))  # all rows of equal length

        @property
        def rows(self):
                return len(self)

        @property
        def cols(self):
                return len(self[0])

        @classmethod
        def from_PBM(cls, file_name):
                """Read a Portable BitMap file; The format of a pbm file is
                P4
                141 141
                lots of binary
                """
                lines = Path(file_name).read_bytes().splitlines()
                code, dimensions, content = lines
                width, height = dimensions.split()
                width, height = int(width), int(height)
                # bitmaps are stored in bytes
                # so the width must be rounded up to a multiple of 8 bits
                width = next((x for x in range(width, width+8) if x%8==0))
                # use format 0=8b to convert each byte to something like 01101111 as a string
                bits = (bit for byte in content for bit in format(byte, '0=8b'))
                # make this list of bits an array with the right dimensions
                l = [[int(next(bits)) for col in range(width)] for row in range(height)]
                return cls(l)


        def raw_print(self):
                for row in self:
                        print(''.join(row))
        

        @staticmethod
        def grouped_into_pairs(l):
                """
                >>> Bitmap.grouped_into_pairs([1,2,3,4,5,6])
                [(1, 2), (3, 4), (5, 6)]
                """
                return list(zip(l[::2], l[1::2]))

        @staticmethod
        def transposed(m):
                """
                >>> Bitmap.transposed([[1,2], [3,4]])
                [[1, 3], [2, 4]]
                """
                return [list(col) for col in zip(*m)]

        def has_even_rows_and_cols(self):
                """
                >>> Bitmap([[0,1,0,1],[1,1,0,1]]).has_even_rows_and_cols()
                True
                """
                return self.rows % 2 == 0 and self.cols % 2 == 0

        def group_quadrants(self):
                """
                >>> Bitmap([[0,1,0,1],[1,1,0,1]]).group_quadrants()
                [[((0, 1), (1, 1)), ((0, 1), (0, 1))]]
                """
                assert self.has_even_rows_and_cols()
                m = self
                m = [self.grouped_into_pairs(row) for row in m]
                m = self.transposed(m)
                m = [self.grouped_into_pairs(col) for col in m]
                m = self.transposed(m)
                return m

        def pad(self):
                """pad array so that we have even number of rows and cols"""
                if self.rows % 2: # odd number of rows
                        self.append([0]*self.cols)
                if self.cols % 2: # odd number of cols
                        m = self
                        m = self.transposed(m)
                        m.append([0]*self.rows)
                        m = self.transposed(m)
                        self[:] = m
                assert self.has_even_rows_and_cols()
                

        def to_unicode(self):
                """
                >>> Bitmap([[0,1,0,1],[1,1,0,1]]).to_unicode()
                '▟▐'
                """
                if not self.has_even_rows_and_cols():
                        self.pad()
                m = self.group_quadrants()
                m = [[quads_to_unicode[q] for q in row] for row in m]
                m = [''.join(row) for row in m]
                m = '\n'.join(m)
                return m


if __name__ == '__main__':
        import doctest
        doctest.testmod()




import sqlite3
conn = sqlite3.connect(":memory:")

def now():
        return conn.execute("select julianday('now','localtime') + 0.5").fetchone()[0]
def today():
        return int(now())

class JulianDate(float):
        
        def __str__(self):
            return conn.execute("select datetime(? - 0.5)", (float(self),)).fetchone()[0][:10]




from vinca._card import Card
from vinca._browser import Browser
from vinca._lib import ansi
from vinca._lib.readkey import readkey
from vinca._lib.julianday import today
from vinca._statistics import Statistics

from rich import print

class Cardlist(list):
        """"""
        ''' A Cardlist is basically just an SQL query linked to a database
        The filter, sort, findall, and slice methods build up this query
        When used as an iterator it is just a list of cards (ids) 
        It is responsible for knowing its database, usually ~/cards.db '''

        def __init__(self, cursor, conditions=['TRUE'],
            ORDER_BY = ' ORDER BY (select max(date) from reviews where reviews.card_id=cards.id) DESC'):
                self._cursor = cursor   
                self._conditions = conditions
                self._ORDER_BY = ORDER_BY
                # we intentionally skip calling super().__init__ because we only are calling ourselves a list
                # so that Fire thinks we are one. We don't actually care about inheritance!

        # overwrite __dir__ so that we fool python FIRE
        # we don't want methods inherited from list
        # to show up in our help message
        def __dir__(self):
                members = super().__dir__()
                hidden = ['extend','index','pop','mro','remove','append',
                          'clear','insert','reverse','copy']
                return [m for m in members if m not in hidden]

        def _copy(self):
                # create a copy of the conditions list obejct (lists are mutable!)
                # we don't want to be affected by subsequent changes to conditions
                return self.__class__(
                        self._cursor,
                        self._conditions.copy(),
                        self._ORDER_BY)

        @property
        def _SELECT_IDS(self):
                """SQL Query of all card IDs. Check this first when debugging."""
                return 'SELECT id FROM cards' + self._WHERE + self._ORDER_BY

        @property
        def _WHERE(self):
                return ' WHERE ' + ' AND '.join(self._conditions)

        def explicit_cards_list(self, LIMIT = 1000):
                self._cursor.execute(self._SELECT_IDS + f' LIMIT {LIMIT}')
                ids = [row[0] for row in self._cursor.fetchall()]
                return [Card(id, self._cursor) for id in ids]

        def __getitem__(self, arg):
                # access cards by their index
                # we use human-oriented indexing beginning with 1...
                if type(arg) is slice:
                        idx = arg.stop 
                elif type(arg) is int:
                        idx = arg
                else:
                        raise ValueError
                self._cursor.execute(self._SELECT_IDS + f' LIMIT 1 OFFSET {idx - 1}')
                card_id = self._cursor.fetchone()[0]
                return Card(card_id, self._cursor)

        def __iter__(self):
                # iterating is discouraged
                # there is usually a better way to do what you want in sql
                # but sometimes this is quick and convenient
                return (self[i] for i in range(1, len(self) + 1))


        def __len__(self):
                self._cursor.execute(f'SELECT COUNT(*) FROM ({self._SELECT_IDS})')
                return self._cursor.fetchone()[0]

        def __str__(self):
                sample_cards = self.explicit_cards_list(LIMIT=6)
                l = len(self)
                s = 'No cards.' if not l else f'6 of {l}\n' if l>6 else ''
                s += ansi.codes['line_wrap_off']
                s += '\n'.join([str(card) for card in sample_cards])
                s += ansi.codes['line_wrap_on']
                return s

        def browse(self):
                """interactively manage you collection"""
                Browser(self.explicit_cards_list(), self._make_basic_card, self._make_verses_card).browse()

        def review(self):
                """review your cards"""
                due_cards = self.filter(due = True).explicit_cards_list()
                Browser(due_cards, self._make_basic_card, self._make_verses_card).review()

        def tag(self, *tags):
                """add tags to selected cards"""
                for tag in tags:
                        self._cursor.execute(f'INSERT INTO tags (card_id, tag)'
                            f'SELECT id, "{tag}" from CARDS' + self._WHERE)
                self._cursor.connection.commit()
                print(self._cursor.rowcount, 'tags added')
                                
        def remove_tag(self, tag):
                """remove a tag from cards """
                self._cursor.execute(f'DELETE FROM tags WHERE card_id IN'
                        f'({self._SELECT_IDS}) AND tag = ?', (tag,))
                self._cursor.connection.commit()
                print(self._cursor.rowcount, 'tags removed')

        def tags(self):
                """all tags in this cardlist """
                self._cursor.execute(f'SELECT tag FROM tags JOIN '
                    f'({self._SELECT_IDS}) ON tags.card_id=id GROUP BY tag')
                return [row[0] for row in self._cursor.fetchall()]

        def count(self):
                """simple summary statistics"""
                return {'total':  len(self),
                        'due':    len(self.filter(due=True)),
                        'new':    len(self.filter(new=True))}

        def filter(self, *,
                   tag = None,
                   created_after=0, created_before=0,
                   due_after=0, due_before=0,
                   deleted=False, due=False, new=False, verses=False,
                   invert=False):
                """filter the collection"""

                TODAY = today() # today's juliandate as an int e.g. 2457035

                parameters_conditions = (
                        # tag
                        (tag, f"(SELECT true FROM tags WHERE card_id=cards.id)"),
                        # date conditions
                        (created_after, f"create_date > {TODAY + created_after}"),
                        (created_before, f"create_date < {TODAY + created_before}"),
                        (due_after, f"due_date > {TODAY + due_after}"),
                        (due_before, f"due_date < {TODAY + due_before}"),
                        # boolean conditions
                        (due, f"due_date < {TODAY}"),
                        (deleted, f"deleted = 1"),
                        (verses, f"verses = 1"),
                        (new, f"due_date = create_date"),
                )

                # assert that at least one filter predicate has been specified
                if not any(parameters_conditions):
                        print('Examples:\n'
                            'filter --new                  new cards\n'
                            'filter --created-after -7     created in the last week\n'
                            '\n'
                            'Read `filter --help` for a complete list of predicates')
                        return

                new_cardlist = self._copy()
                for parameter, condition in parameters_conditions:
                        if parameter:
                                n = 'NOT ' if invert else ''
                                new_cardlist._conditions.append(n + condition)
                return new_cardlist

        def find(self, pattern):
                """ return the first card containing a search pattern """
                try:
                        return self.findall(pattern).sort('seen')[1]
                except:
                        return f'no cards containing "{pattern}"'


        def findall(self, pattern):
                """ return all cards containing a search pattern """
                new_cardlist = self._copy()
                new_cardlist._conditions += [f"(front_text LIKE '%{pattern}%' \
                        OR back_text LIKE '%{pattern}%')"]
                return new_cardlist

        def sort(self, criterion=None, *, reverse=False):
                """ sort the collection: [due | seen | created | time | random] """
                # E.g. we want to see the cards that have taken the most time first
                        
                crit_dict = {'due': ' ORDER BY due_date',
                             'created': ' ORDER BY create_date',
                             'random': ' ORDER BY RANDOM()',
                             'time': ' ORDER BY (select sum(seconds) from reviews where reviews.card_id=cards.id)',
                             'seen': ' ORDER BY (select max(date) from reviews where reviews.card_id=cards.id)',
                              }
                if criterion not in crit_dict:
                        print(f'supply a criterion: {" | ".join(crit_dict.keys())}')
                        exit()
                new_cardlist = self._copy()
                new_cardlist._ORDER_BY = crit_dict[criterion]
                # Sometimes it is natural to see the highest value first by default
                reverse ^= criterion in ('created', 'seen', 'time') 
                direction = ' DESC' if reverse else ' ASC'
                new_cardlist._ORDER_BY += direction
                return new_cardlist
                

        def time(self):
                """ total time spend studying these cards """
                self._cursor.execute(f'SELECT sum(seconds) FROM ({self._SELECT_IDS})'
                        ' LEFT JOIN reviews ON id=card_id')
                minutes = self._cursor.fetchone()[0] // 60
                hours, minutes = divmod(minutes, 60)
                return f'{hours:0>2d}:{minutes:0>2d}'

        def _make_basic_card(self):
                """ make a basic question and answer flashcard """
                card = Card._new_card(self._cursor)
                card.edit()
                return card
        basic = _make_basic_card

        def _make_verses_card(self):
                """ make a verses card: for recipes, poetry, oratory, instructions """
                card = Card._new_card(self._cursor)
                card.verses = True
                card.edit()
                return card
        verses = _make_verses_card

        def delete(self):
                print(f'[bold]delete {len(self)} cards? y/n')
                if readkey() != 'y':
                        print('[red]aborted')
                        return
                self._cursor.execute(f'UPDATE cards SET deleted = 1' + self._WHERE)
                self._cursor.connection.commit()
                print(len(self), 'cards deleted')

        def restore(self):
                deleted_cards = self.filter(deleted=True)
                if len(deleted_cards) == 0:
                        print('none of these cards are deleted')
                self._cursor.execute(f'UPDATE cards SET deleted = 0' + deleted_cards._WHERE)
                self._cursor.connection.commit()
                print(self._cursor.rowcount, 'cards restored')

        def purge(self):
                """ permanently remove deleted cards """
                deleted_cards = self.filter(deleted=True)
                n = len(deleted_cards)
                if n == 0:
                        print('no cards to delete')
                        return
                print(f'[bold]permanently remove {len(deleted_cards)} cards?! y/n')
                if readkey() != 'y':
                        print('[red]aborted')
                        return
                deleted_cards._cursor.execute("DELETE FROM cards" + deleted_cards._WHERE)
                print(f'{deleted_cards._cursor.rowcount} cards purged')
                deleted_cards._cursor.connection.commit()

        def stats(self, interval=7):
                """ review statistics for the collection """
                return Statistics(self._cursor, interval=interval).print()



from vinca._lib import ansi
from vinca._lib.terminal import LineWrapOff, AlternateScreen
from vinca._lib.readkey import readkey, keys

FRAME_WIDTH = 6


class Browser:
    quit_keys = ('q', keys.ESC)
    move_keys = ('j', 'k', keys.DOWN, keys.UP)

    def __init__(self, explicit_cardlist, make_basic_card, make_verses_card):
        self.cardlist = explicit_cardlist
        self.reviewing = False
        self.sel = 0
        self.frame = 0
        self.make_basic_card = make_basic_card
        self.make_verses_card = make_verses_card

    def __len__(self):
        return len(self.cardlist)

    @property
    def selected_card(self):
        return self.cardlist[self.sel]

    @property
    def visible_lines(self):
        return min(len(self), FRAME_WIDTH) + bool(self.status_bar)

    @property
    def status_bar(self):
        bar_text = ansi.codes['light'] + f'{self.sel + 1} of {len(self)}.' + \
                   '  ? for help\n' + ansi.codes['reset']
        return bar_text if len(self) > FRAME_WIDTH else ''

    def draw_browser(self):
        ansi.hide_cursor()
        with LineWrapOff():
            print(self.status_bar, end='')
            visible_cards = self.cardlist[self.frame:self.frame + FRAME_WIDTH]
            for i, card in enumerate(visible_cards, start=self.frame):
                if card.is_due:
                    ansi.blue()
                if card.deleted:
                    ansi.red()
                if i == self.sel:
                    ansi.highlight()
                print(card)
                ansi.reset()

    def clear_browser(self):
        ansi.up_line(self.visible_lines)
        ansi.clear_to_end()

    def close_browser(self):
        self.clear_browser()
        ansi.show_cursor()
        exit()

    def redraw_browser(self):
        self.clear_browser()
        self.draw_browser()

    def move(self, key):
        if key in ('j', keys.DOWN):
            self.move_down()
        if key in ('k', keys.UP):
            self.move_up()

    def move_down(self):
        if self.sel == len(self) - 1:
            return  # we are already at the bottom
        self.sel += 1
        # scroll down if we are off the screen
        self.frame += (self.frame + FRAME_WIDTH == self.sel)

    def move_up(self):
        if self.sel == 0:
            return  # we are already at the top
        self.sel -= 1
        # scroll up if we are off the screen
        self.frame -= (self.frame - 1 == self.sel)

    def print_help(self):
        self.clear_browser()
        ansi.show_cursor()
        print(''
              'J      move down               \n'
              'K      move up                 \n'
              'R      review                  \n'
              'E      edit                    \n'
              'T      edit tags               \n'
              'Q      quit                    \n'
              )
        exit()

    def review(self):
        self.reviewing = True
        self.browse()

    def browse(self):
        if not self.cardlist:
            print('no cards')
            return
        self.draw_browser()

        while True:
            self.redraw_browser()

            if self.reviewing:
                self.selected_card.review()
                if self.selected_card.last_action_grade == 'exit':
                    # exit reviewing mode
                    self.reviewing = False
                    continue
                # close the browser if we have reached the bottom
                if self.sel == len(self) - 1:
                    self.close_browser()
                # move down to the next card
                self.move_down()
                continue

            k = readkey()

            if k == 'r' or k == '\n' or k == '\r':
                self.reviewing = True

            if k == '?':
                self.print_help()

            if k in self.quit_keys:
                self.close_browser()

            if k in self.move_keys:
                self.move(k)

            if command := self.selected_card._hotkeys.get(k):
                with AlternateScreen():
                    command()

            if k in ('b', 'v'):
                with AlternateScreen():
                    new_card = self.make_basic_card() if k == 'b' \
                        else self.make_verses_card() if k == 'v' else None
                    self.cardlist.insert(self.sel, new_card)
                # if this makes us draw a status bar we go down an extra line
                if len(self) == FRAME_WIDTH + 1:
                    ansi.down_line()



"""Spaced Repetition CLI"""

import sqlite3 as _sqlite3
from vinca._cardlist import Cardlist as _Cardlist
from vinca._config import collection_path, __file__
from pathlib import Path as _Path

from rich import print as _print


collection_path = _Path(collection_path).expanduser()
_vinca_path = _Path(__file__).parent
_empty_deck_path = _vinca_path / 'empty_deck.db'
_tutorial_path = _vinca_path / 'tutorial_cards.db'

# create a collection if it does not exist
if not collection_path.exists():
        print(f'no collection found at {collection_path}')
        import shutil
        shutil.copy(_empty_deck_path, collection_path)
        print('empty collection created')

# create default collection
_cursor = _sqlite3.connect(collection_path).cursor()
col = collection = _Cardlist(_cursor)

# The "tutorial" is just a deck of cards used to teach the basics of vinca
_tutorial_cursor = _sqlite3.connect(_tutorial_path).cursor()
tutorial = _Cardlist(_tutorial_cursor)

# import some methods of the collection Cardlist object directly into the module's namespace
# this is so that ```vinca col review``` can be written as ```vinca review```
_methods = ('browse', 'count', 'filter', 'find', 'findall', 'review', 'sort', 'time', 'purge', 'basic', 'verses', 'stats')
for _method_name in _methods:
    globals()[_method_name] = getattr(collection, _method_name)

globals()['1'] = lambda: collection[1]
globals()['1'].__doc__ = "most recent card"
globals()['2'] = lambda: collection[2]
globals()['2'].__doc__ = "second most recent card"
globals()['3'] = lambda: collection[3]
globals()['3'].__doc__ = "third most recent card"


def edit_config():
    from subprocess import run
    run(['vim', __file__])


def help():
    """print basic help"""
    _print('\n',
           '[bold green] --help                ', 'full screen help                        \n',
           '[bold green] basic                 ', 'create question and answer cards        \n',
           '[bold green] review                ', 'study your cards                        \n',
           '[bold green] browse                ', 'interactively manage your cards         \n',
           '[bold green] count                 ', 'simple summary statistics               \n',
           '[bold green] tutorial review       ', 'study a tutorial deck of twenty cards   \n', sep='')
globals()['-h'] = help



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vinca', 'vinca._lib']

package_data = \
{'': ['*']}

install_requires = \
['fire>=0.4.0,<0.5.0', 'prompt-toolkit>=3.0.29,<4.0.0', 'rich>=12.4.1,<13.0.0']

entry_points = \
{'console_scripts': ['vinca = vinca:run']}

setup_kwargs = {
    'name': 'vinca',
    'version': '2.3.1',
    'description': 'Spaced Repetition CLI',
    'long_description': '## Synopsis\n\n`VINCA` is a command-line spaced-repetition study tool.\n\n## Installing Vinca\n\n`pip install vinca`\n\n## Basic commands\n\n|command           |   description                            |  \n|------------------|------------------------------------------|  \n|--help            |   full screen help                       |  \n|basic             |   create question and answer cards       |  \n|review            |   study your cards                       |  \n|browse            |   interactively manage your cards        |  \n|count             |   simple summary statistics              |  \n|tutorial review   |   study a tutorial deck of twenty cards  |  \n\n## Screenshots\n\n![vinca stats](./utils/stats.png)\n*`vinca stats`*\n\n![test](./utils/screencast.gif?)\n*test*\n',
    'author': 'Oscar Laird',
    'author_email': 'olaird25@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/oscarlaird/vinca-SRS',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import time
from pathlib import Path

from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter

from rich import print

from vinca._lib.terminal import AlternateScreen
from vinca._lib.readkey import readkey
from vinca._lib.video import DisplayImage
from vinca._lib import ansi
from vinca._lib import julianday

TODAY = julianday.today()  # int representing day

GRADE_DICT = {'x': 'delete', 'd': 'delete',
              'q': 'exit', '\x1b': 'exit',
              'p': 'preview', '0': 'preview',
              '1': 'again',
              '2': 'hard',
              '3': 'good', ' ': 'good', '\r': 'good', '\n': 'good',
              '4': 'easy'}
RESET_ACTION_GRADES = ('create', 'again')
STUDY_ACTION_GRADES = ('hard', 'good', 'easy')
BUREAU_ACTION_GRADES = ('edit', 'delete', 'exit', 'preview')

help_string = ('[dim]'
               'Q  [white]quit[/white]                   \n'
               'E  [white]edit[/white]                   \n'
               'D  [red]delete[/red]      \n'
               '1  [red]again[/red]       \n'
               '2  [red]hard[/red]        \n'
               '3  [blue]good[/blue]      \n'
               '4  [green]easy[/green]    \n')


class Card:
    # A card is a dictionary
    # its data is loaded from SQL on the fly and saved to SQL on the fly

    _bool_fields = ('deleted', 'verses')
    _date_fields = ('create_date', 'due_date')
    _text_fields = ('front_text', 'back_text')
    _BLOB_fields = ('front_image', 'back_image', 'front_audio', 'back_audio')
    _media_fields = _text_fields + _BLOB_fields
    _fields = ('id',) + _date_fields + _media_fields + _bool_fields

    # let us access key-val pairs from the dictionary as simple attributes
    for f in _fields:
        exec(
            f'''
@property
def {f}(self):
        return self["{f}"]
@{f}.setter
def {f}(self, new_value):
        self["{f}"] = new_value
'''
        )
    del f

    def __init__(self, id, cursor):
        self._cursor = cursor
        self._dict = dict(id=id)
        self._hotkeys = {'e': self.edit,
                         't': self.edit_tags,
                         'd': self._toggle_delete,
                         '+': self.postpone, }

    def __str__(self):
        s = ''
        if self.deleted:
            s += ansi.codes['red']
        elif self.is_due:
            s += ansi.codes['blue']
        s += self.front_text.replace('\n', ' / ')
        s += ' | '
        s += self.back_text.replace('\n', ' / ')
        s += ansi.codes['reset']
        return s

    def metadata(self):
        return {field: str(getattr(self, field)) for field in self._fields +
                ('interval', 'ease', 'last_study_date', 'last_reset_date', 'tags')}

    @staticmethod
    def _is_path(arg):
        """ Check if an argument specifies a file """
        if not arg or type(arg) not in (str, Path):
            return
        try:
            return Path(arg).exists()
        except:
            return

    def __setitem__(self, key, value):
        assert key != 'id', 'Card Id cannot be changed!'
        assert key in self._fields
        # if the supplied value is a filename we want
        # to use the contents of the file, not the filename itself
        if key in self._media_fields and self._is_path(value):
            if key in self._text_fields:
                value = Path(value).read_text()
            if key in self._BLOB_fields:
                value = Path(value).read_bytes()
        self._dict[key] = value
        # commit change to card-dictionary to SQL database
        self._cursor.execute(f'UPDATE cards SET {key} = ? WHERE id = ?',
                             (value, self.id))
        self._cursor.connection.commit()

    set = __setitem__  # alias for easy cli usage
    set.__doc__ = 'set the text or image for a card: `set front_image ./diagram.png`'

    def __getitem__(self, key):
        if key not in self._fields:
            raise KeyError(f'Field "{key}" does not exist')
        if key not in self._dict.keys():
            # load attribute from the database if we haven't yet
            value = self._cursor.execute(f'SELECT {key} FROM cards'
                                         ' WHERE id = ?', (self.id,)).fetchone()[0]
            # preprocess certain values to cast them to better types:
            if key in self._text_fields:
                # if SQL passes us an Integer or None
                # this is going to cause errors
                value = str(value or '')
            if key in self._date_fields:
                value = julianday.JulianDate(value)
            self._dict[key] = value
        return self._dict[key]

    def __len__(self):
        # Fire sees __getitem__ and thinks we can be indexed
        # by defining len=0 we tell it not to try to index us
        return 0

    def delete(self):
        self.deleted = True
        return 'to undo use the restore command'

    def _toggle_delete(self):
        self.deleted = not self.deleted

    def restore(self):
        self.deleted = False
        return 'card restored'

    @property
    def is_due(self):
        return self.due_date <= TODAY

    def postpone(self, n=1):
        # 'Make card due n days after today. (default 1)'
        tomorrow = TODAY + n
        hour = self.due_date % 1
        self.due_date = tomorrow + hour
        return f'Postponed until {self.due_date}.'

    def review(self):
        start = time.time()
        grade_key = self._review_verses() if self.verses else self._review_basic()
        grade = GRADE_DICT.get(grade_key, 'exit')
        stop = time.time()

        elapsed_seconds = int(stop - start)

        self._log(grade, elapsed_seconds)
        if self.last_action_grade == 'delete':
            self.deleted = True
        self._schedule()

    def _review_basic(self):
        def edit_then_review():
            ansi.move_to_top();
            ansi.clear_to_end()
            self.edit()
            return self._review_basic()

        with AlternateScreen():
            print(f'[bold]{self.front_text}')
            print('\n', f'[dim yellow italic]{" ".join(self.tags)}', '\n', sep='')
            with DisplayImage(data_bytes=self.front_image):
                char = readkey()  # press any key to flip the card
                if char == 'e':  # edit the card and then review it
                    return edit_then_review()
                if char in ['d', 'q']:  # immediate exit actions
                    return char
            with DisplayImage(data_bytes=self.back_image):
                print(f'[bold]{self.back_text}')
                print('\n\n')
                print(help_string)
                grade = readkey()
                if grade == 'e':
                    return edit_then_review()
                return grade

    def _review_verses(self):
        def edit_then_review():
            ansi.move_to_top();
            ansi.clear_to_end()
            self.edit()
            return self._review_verses()

        with AlternateScreen():
            ansi.light();
            print(self.tags, '\n', sep='');
            ansi.reset()
            lines = self.front_text.splitlines()
            print(f'[bold]{lines.pop(0)}')
            for line in lines:
                char = readkey()  # press any key to continue
                if char == 'e':
                    return edit_then_review()
                if char in ['d', 'q']:  # exit actions to abort the review
                    return char
                print(f'[bold]{line}')

            # grade the card
            print('\n\n[END]')
            print('\n\n')
            print(help_string)
            grade = readkey()
            if grade == 'e':
                return edit_then_review()
            return grade

    def edit(self):
        start = time.time()
        self._edit_verses() if self.verses else self._edit_basic()
        stop = time.time()
        elapsed_seconds = int(stop - start)
        self._log('edit', elapsed_seconds)

    def _edit_basic(self):
        self.front_text = prompt('Question:     ',
                                 default=self.front_text,
                                 multiline=True, vi_mode=True,
                                 bottom_toolbar=lambda: 'press ESC-Enter to confirm')
        self.back_text = prompt('Answer:       ',
                                default=self.back_text,
                                multiline=True, vi_mode=True,
                                bottom_toolbar=lambda: 'press ESC-Enter to confirm')

    def _edit_verses(self):
        self.front_text = prompt('Question:     ',
                                 default=self.front_text,
                                 multiline=True, vi_mode=True,
                                 bottom_toolbar=lambda: 'press ESC-Enter to confirm')

    def _log(self, action_grade, seconds):
        self._cursor.execute('INSERT INTO reviews (card_id, seconds, action_grade)'
                             ' VALUES (?, ?, ?)', (self.id, seconds, action_grade))
        self._cursor.connection.commit()

    @property
    def last_action_grade(self):
        return self._cursor.execute('SELECT action_grade, max(date)'
                                    ' FROM reviews WHERE card_id = ?', (self.id,)).fetchone()[0]

    @property
    def last_reset_date(self):
        date = self._cursor.execute('SELECT max(date)'
                                    ' FROM reviews WHERE action_grade IN (?, ?) AND card_id = ?',
                                    RESET_ACTION_GRADES + (self.id,)).fetchone()[0]
        if date is None:
            return self.create_date
        return julianday.JulianDate(date)

    @property
    def last_study_date(self):
        date = self._cursor.execute('SELECT max(date)'
                                    ' FROM reviews WHERE action_grade IN (?, ?, ?, ?, ?) AND card_id = ?',
                                    RESET_ACTION_GRADES + STUDY_ACTION_GRADES + (self.id,)).fetchone()[0]
        if date is None:
            return self.create_date
        return julianday.JulianDate(date)

    @property
    def ease(self):
        # the ease is the ratio of the card's age to the next interval
        # if ease=1 the intervals will double
        # we calculate ease as an average of our grades
        # consistently grading 'good' yields ease=1
        grades_since_reset = [grade[0] for grade in \
                              list(self._cursor.execute('SELECT action_grade FROM reviews'
                                                        ' WHERE card_id = ? AND date > ?'
                                                        ' AND action_grade IN (?, ?, ?)',
                                                        (self.id,
                                                         self.last_reset_date) + STUDY_ACTION_GRADES).fetchall())
                              ]
        grade_ease_dict = {'hard': 0, 'good': 1, 'easy': 2}
        e = [grade_ease_dict[grade] for grade in grades_since_reset]
        if not e:
            return 1
        return sum(e) / len(e)

    @property
    def interval(self):
        # The interval for the next review is calculated from two values:
        # ✠ The Ease
        # ✠ The number of days between creation (or reset) and the most recent study
        #   This is called "study maturity"
        study_maturity = int(self.last_study_date) - int(self.last_reset_date)
        interval = int(self.ease * study_maturity)
        return max(1, interval)

    def _calc_due_date(self):
        if self.last_action_grade in RESET_ACTION_GRADES:
            return self.last_reset_date
        return self.last_study_date + self.interval

    def _schedule(self):
        self.due_date = self._calc_due_date()

    @property
    def tags(self):
        # raise ValueError(str(self.id))
        tags = self._cursor.execute('SELECT tag FROM tags'
                                    ' WHERE card_id = ?', (self.id,)).fetchall()
        tags = [row[0] for row in tags]
        return tags

    def _remove_tag(self, tag):
        self._cursor.execute('DELETE FROM tags'
                             ' WHERE card_id = ? AND tag = ?', (self.id, tag))
        self._cursor.connection.commit()

    def _add_tag(self, tag):
        if tag not in self.tags:
            self._cursor.execute('INSERT INTO tags'
                                 ' (card_id, tag) VALUES (?, ?)', (self.id, tag))
            self._cursor.connection.commit()

    def edit_tags(self):
        new_tags = prompt('tags: ',
                          default=' '.join(self.tags),
                          completer=WordCompleter(self._collection_tags()),
                          ).split()
        for tag in self.tags:
            self._remove_tag(tag)
        for tag in new_tags:
            self._add_tag(tag)

    def _collection_tags(self):
        self._cursor.execute('SELECT tag FROM tags GROUP BY tag')
        tags = [row[0] for row in self._cursor.fetchall()]
        return tags

    def tag(self, *tags):
        """ add tags to the card """
        for tag in tags:
            self._add_tag(tag)

    @classmethod
    def _new_card(cls, cursor):
        cursor.execute("INSERT INTO cards DEFAULT VALUES")
        cursor.connection.commit()
        id = cursor.execute("SELECT last_insert_rowid()").fetchone()[0]
        return cls(id, cursor)



# A simple module for reading a keystroke from the user
# adapted from the getchar function in the click library

import tty
import termios
import sys
import contextlib
import os
from types import SimpleNamespace

keys = SimpleNamespace( UP = '\x1b[A', DOWN = '\x1b[B', RIGHT = '\x1b[C', LEFT = '\x1b[D',
                        CTRL_UP = '\x1b[1;5A', CTRL_DOWN = '\x1b[1;5B', CTRL_RIGHT = '\x1b[1;5C', CTRL_LEFT = '\x1b[1;5D',
                        ESC = '\x1b', BACK = '\x7f', CTRL_R = '\x12', CTRL_K = '\x0b')

@contextlib.contextmanager
def raw_terminal():
        fd = sys.stdin.fileno()

        old_settings = termios.tcgetattr(fd) # save state of terminal

        try:
                # the tty has two modes:
                # 'cooked' waits for the user to press enter
                # 'raw' passes the typed characters immediately
                tty.setraw(fd)
                yield fd
        finally:
                # restore state of terminal
                termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
                sys.stdout.flush() 


def readkey():
        with raw_terminal() as fd:
                # a single key can be represented by up to 4 bytes
                return os.read(fd, 6).decode()
